/*
 * ----------------------------------------------------------------------------
 *  (C) Copyright Testing Technologies, 2001-2005.  All Rights Reserved.
 *
 *  All copies of this program, whether in whole or in part, and whether
 *  modified or not, must display this and all other embedded copyright
 *  and ownership notices in full.
 *
 *  See the file COPYRIGHT for details of redistribution and use.
 *
 *  You should have received a copy of the COPYRIGHT file along with
 *  this file; if not, write to the Testing Technologies,
 *  Rosenthaler Str. 13, 10119 Berlin, Germany.
 *
 *  TESTING TECHNOLOGIES DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS
 *  SOFTWARE. IN NO EVENT SHALL TESTING TECHNOLOGIES BE LIABLE FOR ANY
 *  SPECIAL, DIRECT, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 *  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN
 *  AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
 *  ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF
 *  THIS SOFTWARE.
 *
 *  THE SOFTWARE IS PROVIDED "AS IS" WITHOUT WARRANTY OF ANY KIND,
 *  EITHER EXPRESSED OR IMPLIED, INCLUDING ANY KIND OF IMPLIED OR
 *  EXPRESSED WARRANTY OF NON-INFRINGEMENT OR THE IMPLIED WARRANTIES
 *  OF MECHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
 *
 * ----------------------------------------------------------------------------
 *
 *  AUTHOR:      Dirk Tepelmann
 *  DATE:        January, 2005
 *
 *  REVISION INFO:
 *        $Revision: 1.6 $ $Date: 2005/09/13 08:47:39 $
 * ----------------------------------------------------------------------------- */
package com.testingtech.ttcn.tri;

import java.math.BigInteger;
import java.net.Inet6Address;
import java.net.InetAddress;
import java.net.UnknownHostException;

import org.etsi.ttcn.tci.CharstringValue;
import org.etsi.ttcn.tci.EnumeratedValue;
import org.etsi.ttcn.tci.IntegerValue;
import org.etsi.ttcn.tci.OctetstringValue;
import org.etsi.ttcn.tci.RecordOfValue;
import org.etsi.ttcn.tci.RecordValue;
import org.etsi.ttcn.tci.UnionValue;
import org.etsi.ttcn.tci.Value;
import org.etsi.ttcn.tri.TriMessage;

import com.testingtech.generated.codec.Ipv6Address;
import com.testingtech.generated.codec.Ipv6Header;
import com.testingtech.generated.codec.MipHeader;
import com.testingtech.generated.codec.Oct1to255;
import com.testingtech.ttcn.tci.codec.tabular.TabularEncoder;
import com.testingtech.ttcn.tci.codec.tabular.TabularException;
import com.testingtech.util.BitArrayOutputStream;

import de.tu_berlin.cs.uebb.muttcn.runtime.RB;
import de.tu_berlin.cs.uebb.ttcn.runtime.OctetString;

/**
 * 
 * This class implements all the external function declaration of the
 * IPv6 ATS(STF276)
 *  
 */
public class ExternalFunctions_IPv6 implements ExternalFunctionsDefinition {
	
	private RB RB;
	
	private final int HOP = 0;
	private final int TUNNELEDIPV6 = 41;
	private final int ROUTING = 43;
	private final int FRAGMENT = 44;
	private final int DESTINATION = 60;
    
	private final int ipv6HeaderSize = 40;

	public ExternalFunctions_IPv6(RB rb)
	{
		RB = rb;
	}
	
	/** 
	 * This external function returns the mac address which corresponds 
	 * to the ip address
	 * 
	 * @param   p_ipv6Address	The given IPv6 address
	 * @return  				Oct1to255, the MAC address
	 */
	public synchronized Oct1to255 fx_getMacAddr(Ipv6Address p_ipv6Address) {
		//TODO???
		Oct1to255 ret = new Oct1to255(RB.getTciCDRequired().getTypeForName("Oct1to255").newInstance());
	    return ret;
	}
	
	/**
	 * This external function maps ports to mac-addresses
	 * 
	 * @param   p_portCfg	The port configuration
	 * @return  			FncRetCode, the status of the external function
	 */
	public synchronized EnumeratedValue fx_configPorts(RecordValue p_portCfg) {
	    EnumeratedValue FncRetCode = (EnumeratedValue)RB.getTciCDRequired().getTypeForName("LibCommon_VerdictControl.FncRetCode").newInstance();
	    FncRetCode.setEnum("e_success");
	    return FncRetCode;
	}

	/**
	 * This external function calculates the payload length 
	 * of a IPv6 packet
	 * 
	 * @param   p_msg 	The IPv6 packet for calculation.
	 * @return  		The payload length.
	 */
	public synchronized IntegerValue fx_payloadLength(UnionValue p_msg ){
	    
	    TriMessage message = RB.getTciEncoding().getCodec("LibIpv6_Interface").encode(p_msg);
	    
	    int msgLen = message.getEncodedMessage().length - ipv6HeaderSize;
	    
//	    IntegerValue ret = (IntegerValue) RB.getTciCDRequired().getInteger().newInstance();
	    IntegerValue ret = (IntegerValue) RB.getTciCDRequired().getTypeForName("LibCommon_BasicTypesAndValues.UInt16").newInstance();

	    ret.setInt(msgLen);
	    
	    return ret;
	}

	/**
	 * This external function calculates the payload length 
	 * of a IPv6 packet
	 * 
	 * @param   p_msg 	The IPv6 packet for calculation.
	 * @param   p_entry   The indicator for which tunnelled header
	 *                  the payload length should be calculated.
	 * @return  		The payload length.
	 */
	public synchronized IntegerValue fx_tunnelledPayloadLength(UnionValue p_msg, IntegerValue p_entry){
	    
		int v_cnt = 0;
		
	    TriMessage message = RB.getTciEncoding().getCodec("LibIpv6_Interface").encode(p_msg);
	    
	    int msgLen = message.getEncodedMessage().length - ipv6HeaderSize;
	    
	    Value mValue = p_msg.getVariant(p_msg.getPresentVariantName());

	    Ipv6Header ipv6Header = new Ipv6Header(((RecordValue)mValue).getField("ipv6Hdr"));
	    try {
            ipv6Header.encode(new TabularEncoder(new BitArrayOutputStream()));
        } catch (TabularException tex) {
            RB.getTciTMProvided().tciError(tex.getMessage());
            tex.printStackTrace();
            return (IntegerValue)RB.getTciCDRequired().getInteger().newInstance();
        }
	    
	    int nextHeader = ipv6Header.nextHeader.tValue;
        
        if (nextHeader != 135) {
	        RecordOfValue ehrov = (RecordOfValue)((RecordValue)mValue).getField("extHdrList");
	        
	        if (!ehrov.notPresent()) {
                for (int i=0; (i<ehrov.getLength())&&(v_cnt!=p_entry.getInt()); i++) {
	                RecordValue rv = (RecordValue) ((UnionValue)ehrov.getField(i)).getVariant(((UnionValue)ehrov.getField(i)).getPresentVariantName());

	    	        switch (nextHeader) {
	    	        case HOP:
	    	            msgLen -= (((IntegerValue)rv.getField("hdrExtLen")).getInt() * 8 + 8);
	    	            nextHeader = ((IntegerValue)rv.getField("nextHeader")).getInt();
	    	            break;
	    	        case TUNNELEDIPV6:
	    	            msgLen -= ipv6HeaderSize;
	    	            nextHeader = ((IntegerValue)rv.getField("nextHeader")).getInt();
	    	            v_cnt++;
	    	            break;
	    	        case ROUTING:
	    	            msgLen -= (((IntegerValue)rv.getField("hdrExtLen")).getInt() * 8 + 8);
	    	            nextHeader = ((IntegerValue)rv.getField("nextHeader")).getInt();
	    	            break;
	    	        case FRAGMENT:
	    	            msgLen -= 8;
	    	            nextHeader = ((IntegerValue)rv.getField("nextHeader")).getInt();
	    	            break;
	    	        case DESTINATION:
	    	            msgLen -= (((IntegerValue)rv.getField("hdrExtLen")).getInt() * 8 + 8);
	    	            nextHeader = ((IntegerValue)rv.getField("nextHeader")).getInt();
	    	            break;
	    	        }
	            }
	        }
	        
        }

	    IntegerValue ret = (IntegerValue) RB.getTciCDRequired().getTypeForName("LibCommon_BasicTypesAndValues.UInt16").newInstance();

	    if (v_cnt == p_entry.getInt()) {
	    	ret.setInt(msgLen);
	    }
	    else {
            RB.getTciTMProvided().tciError("fx_tunnelledPayloadLength: Error while calculating payload length.\n" +
            		"Reason: Ipv6 packet does not include the amount of tunnelled Ipv6 headers(41)(found " + v_cnt +
            		" occurence(s), requested was " + p_entry.getInt() + " entry)");
	    }
	    
	    return ret;
	}

	/**
	 * This external function calculates the checksum
	 * 
	 * @param  p_packet The Icmpv6 packet for calculation. 
	 * @return 			The calculated checksum.
	 */
	public synchronized OctetstringValue fx_icmpv6Checksum(UnionValue p_packet) {
		
	    byte [] pseudoHeader = new byte[40];
	    
	    int eh_size = 0;
	    
	    int checksum = 0;
	    
	    boolean homeAddressIncluded = false;
	    byte [] homeAddress = null;
	    
	    Value mValue = p_packet.getVariant(p_packet.getPresentVariantName());

	    Ipv6Header ipv6Header = new Ipv6Header(((RecordValue)mValue).getField("ipv6Hdr"));
	    try {
            ipv6Header.encode(new TabularEncoder(new BitArrayOutputStream()));
        } catch (TabularException tex) {
            RB.getTciTMProvided().tciError(tex.getMessage());
            tex.printStackTrace();
            return (OctetstringValue)RB.getTciCDRequired().getOctetstring().newInstance();
        }
	    
	    int nextHeader = ipv6Header.nextHeader.tValue;
        
        if (nextHeader != 58) {
	        RecordOfValue ehrov = (RecordOfValue)((RecordValue)mValue).getField("extHdrList");
	        
	        if (!ehrov.notPresent()) {
                for (int i=0; i<ehrov.getLength(); i++) {
	                RecordValue rv = (RecordValue) ((UnionValue)ehrov.getField(i)).getVariant(((UnionValue)ehrov.getField(i)).getPresentVariantName());

	    	        switch (nextHeader) {
	    	        case HOP:
	    	            eh_size += (((IntegerValue)rv.getField("hdrExtLen")).getInt() * 8 + 8);
	    	            nextHeader = ((IntegerValue)rv.getField("nextHeader")).getInt();
	    	            break;
	    	        case TUNNELEDIPV6:
	    	            eh_size += ipv6HeaderSize;
	    	            nextHeader = ((IntegerValue)rv.getField("nextHeader")).getInt();
	    	            break;
	    	        case ROUTING:
	    	            eh_size += (((IntegerValue)rv.getField("hdrExtLen")).getInt() * 8 + 8);
	    	            nextHeader = ((IntegerValue)rv.getField("nextHeader")).getInt();
	    	            break;
	    	        case FRAGMENT:
	    	            eh_size += 8;
	    	            nextHeader = ((IntegerValue)rv.getField("nextHeader")).getInt();
	    	            break;
	    	        case DESTINATION:
	    	            eh_size += (((IntegerValue)rv.getField("hdrExtLen")).getInt() * 8 + 8);
	    	            nextHeader = ((IntegerValue)rv.getField("nextHeader")).getInt();
	    	            RecordOfValue rov = (RecordOfValue)rv.getField("destOptionList");
	    	            for (int j=0; j<rov.getLength(); j++) {
	    	            	UnionValue destOption = (UnionValue)rov.getField(j);
							Value variant = destOption.getVariant(destOption.getPresentVariantName());
							if (destOption.getPresentVariantName().equals("homeAddressOption")) {
	    	            		homeAddressIncluded = true;
	    	            		OctetstringValue ov = (OctetstringValue)((RecordValue)variant).getField("homeAddress");
	    	            		homeAddress = new OctetString(RB, ov.getString()).toByteArray();
	    	            	}
	    	            }
	    	            break;
	    	        }
	            }
	        }
	        
        }

        /* Due to unknown next header is used too, we skip this check and assume 
         * that this function is only called for ICMP message checksum calculation.
         */
        //if (nextHeader == 58) {
	        //16Byte src
            byte []src = null;
            if (!homeAddressIncluded) {
            	src = ipv6Header.sourceAddress.tValue;
            }
            else {
            	src = homeAddress;
            }
	        System.arraycopy(src, 0, pseudoHeader, 0, 16);
	        //16Byte dst
            byte []dst = ipv6Header.destinationAddress.tValue;
	        System.arraycopy(dst, 0, pseudoHeader, 16, 16);
	        
	        //4Byte upper layer packet length
	        byte[] upperLayerLength = new byte[4];
	        int len = ipv6Header.payloadLength.tValue - eh_size;
	        upperLayerLength[0] = (byte)(len>>24);
	        upperLayerLength[1] = (byte)(len>>16);
	        upperLayerLength[2] = (byte)(len>>8);
	        upperLayerLength[3] = (byte)(len);
	        System.arraycopy(upperLayerLength, 0, pseudoHeader, 32, 4);
	        
	        //3Byte zero
	        byte [] zero_bytes = new byte[] {0x0, 0x0, 0x0};
	        System.arraycopy(zero_bytes, 0, pseudoHeader, 36, 3);
	        
	        //1Byte next header
	        byte [] next_header = new byte[] {0x3A };
	        System.arraycopy(next_header, 0, pseudoHeader, 39, 1);
	        
	        //checksum pseudo header
	        byte []val = new byte[2];
	        for (int i=0; i<40; i+=2) {
	            System.arraycopy(pseudoHeader, i, val, 0, 2);
	            checksum += new BigInteger(val).intValue()&0x0000FFFF;
	            checksum = (checksum>>16) + (checksum & 0x0000FFFF);
	        }
	        
	        //checksum layer 4 header
	        byte [] wholeMessage = RB.getTciEncoding().getCodec("LibIpv6_Interface").encode(p_packet).getEncodedMessage();
	        for (int i=0; i<len; i+=2) {
	        	if ((i+2)<=len) {
	        		System.arraycopy(wholeMessage, i + ipv6HeaderSize + eh_size, val, 0, 2);
	        		checksum += new BigInteger(val).intValue()&0x0000FFFF;
	        		checksum = (checksum>>16) + (checksum & 0x0000FFFF);
	        	}
	        }
	        
	        if ((len%2) != 0) {
	            byte v1 = wholeMessage[len - 1 + ipv6HeaderSize + eh_size];
	            val[0] = v1;
	            val[1] = 0x0;

	            checksum += new BigInteger(val).intValue()&0x0000FFFF;
	            checksum = (checksum>>16) + (checksum & 0x0000FFFF);
	        }
            
            checksum = (~checksum) & 0x0000FFFF;
	        
	    //}
	    
//	    OctetstringValue ret = (OctetstringValue) RB.getTciCDRequired().getOctetstring().newInstance();
	    OctetstringValue ret = (OctetstringValue) RB.getTciCDRequired().getTypeForName("LibCommon_DataStrings.Oct2").newInstance();
	    ret.setLength(2);
	    ret.setOctet(0, (int)checksum>>8);
	    ret.setOctet(1, (int)checksum&0xFF);
	    
	    return ret;
	}
	
	/**
	 * This external function calculates the header length 
	 * of a MIP header.
	 * 
	 * @param   p_hdr 	The MIP header for calculation.
	 * @return  		The header length.
	 */
	//TODO not checked
	public synchronized IntegerValue fx_mipHeaderLength(RecordValue p_hdr ){
	    TriMessage message = RB.getTciEncoding().getCodec("LibIpv6_MultiRfcs_TypesAndValues").encode(p_hdr);
	    
		//  8-bit unsigned integer, representing the length of the Mobility
	    //  Header in units of 8 octets, excluding the first 8 octets.
	    int msgLen = message.getEncodedMessage().length/8-1;

	    IntegerValue ret = (IntegerValue) RB.getTciCDRequired().getTypeForName("LibCommon_BasicTypesAndValues.UInt8").newInstance();

	    ret.setInt(msgLen);
	    
	    return ret;
	}

	/**
	 * This external function calculates the checksum for the mobility header.
	 * 
	 * @param  p_packet The Mipv6 packet for calculation. 
	 * @return 			The calculated checksum.
	 */
	public synchronized OctetstringValue fx_mipHeaderChecksum(OctetstringValue srcAddress, OctetstringValue dstAddress, RecordValue p_hdr) {
	    byte [] pseudoHeader = new byte[40];
	    
	    int checksum = 0;
	    
	    MipHeader mipHeader = new MipHeader(p_hdr);
	    try {
	    	mipHeader.encode(new TabularEncoder(new BitArrayOutputStream()));
        } catch (TabularException tex) {
            RB.getTciTMProvided().tciError(tex.getMessage());
            tex.printStackTrace();
            return (OctetstringValue)RB.getTciCDRequired().getOctetstring().newInstance();
        }
	    
        //16Byte src
//        byte []src = ipv6Header.sourceAddress.tValue;
	    //TODO for debugging reason
        byte []src = ((OctetString)srcAddress).toByteArray();
        System.arraycopy(src, 0, pseudoHeader, 0, 16);
        //16Byte dst
//        byte []dst = ipv6Header.destinationAddress.tValue;
        byte []dst = ((OctetString)dstAddress).toByteArray();
        System.arraycopy(dst, 0, pseudoHeader, 16, 16);
        
        //4Byte upper layer packet length
        byte[] upperLayerLength = new byte[4];
        int len = mipHeader.headerLen.tValue * 8 + 8;
        upperLayerLength[0] = (byte)(len>>24);
        upperLayerLength[1] = (byte)(len>>16);
        upperLayerLength[2] = (byte)(len>>8);
        upperLayerLength[3] = (byte)(len);
        System.arraycopy(upperLayerLength, 0, pseudoHeader, 32, 4);
        
        //3Byte zero
        byte [] zero_bytes = new byte[] {0x0, 0x0, 0x0};
        System.arraycopy(zero_bytes, 0, pseudoHeader, 36, 3);
        
        //1Byte next header
        byte [] next_header = new byte[] {(byte)0x87};//{0x02 };
        System.arraycopy(next_header, 0, pseudoHeader, 39, 1);
        
        //checksum pseudo header
        byte []val = new byte[2];
        for (int i=0; i<40; i+=2) {
        	System.arraycopy(pseudoHeader, i, val, 0, 2);
        	checksum += new BigInteger(val).intValue()&0x0000FFFF;
        	checksum = (checksum>>16) + (checksum & 0x0000FFFF);
        }
        
        //checksum layer 4 header
        byte [] wholeMessage = RB.getTciEncoding().getCodec("LibIpv6_MultiRfcs_TypesAndValues").encode(p_hdr).getEncodedMessage();
        for (int i=0; i<len; i+=2) {
        	if ((i+2)<=len) {
        		System.arraycopy(wholeMessage, i, val, 0, 2);
        		checksum += new BigInteger(val).intValue()&0x0000FFFF;
        		checksum = (checksum>>16) + (checksum & 0x0000FFFF);
        	}
        }
        
        if ((len%2) != 0) {
        	byte v1 = wholeMessage[len - 1];
        	val[0] = v1;
        	val[1] = 0x0;
        	
        	checksum += new BigInteger(val).intValue()&0x0000FFFF;
        	checksum = (checksum>>16) + (checksum & 0x0000FFFF);
        }
        
        checksum = (~checksum) & 0x0000FFFF;
        
	    OctetstringValue ret = (OctetstringValue) RB.getTciCDRequired().getTypeForName("LibCommon_DataStrings.Oct2").newInstance();
	    ret.setLength(2);
	    ret.setOctet(0, (int)checksum>>8);
	    ret.setOctet(1, (int)checksum&0xFF);
	    
	    return ret;
	}
	
	private static final String[] hex = { "0123456789abcdef", "0123456789ABCDEF" };

	/**
	 * This external function returns the encoded IPv6 packet.
	 * 
	 * @param   p_msg 	The IPv6 packet for encoding.
	 * @return  		The encoded packet.
	 */
	public synchronized OctetstringValue fx_encodeMessage(UnionValue p_msg ){
	    
	    TriMessage msg = RB.TestAdapter.getCodec("").encode(p_msg);
	    byte bmsg[] = msg.getEncodedMessage();

		StringBuffer sb = new StringBuffer();
		
		for (int j = 0; j < bmsg.length; j++) {
		    //we choose capital letters
		    sb.append(hex[1].charAt((bmsg[j] >> 4) & 0xf));
		    sb.append(hex[1].charAt((bmsg[j]) & 0xf));
		}		

	    OctetstringValue ret = (OctetstringValue) RB.getTciCDRequired().getOctetstring().newInstance();
	    ret.setLength(bmsg.length);
	    ret.setString("'" + sb + "'O");
	    
	    return ret;
	}

	/** This external function calculates a Message Authentication Code
	 * @param  p_key Key used to compute the MAC
	 * @param  p_message Octetstring message
	 * @param  p_crypto Cryptographic function used to compute MAC
	 * @return Message HMAC
	*/
	public synchronized OctetstringValue fx_computeHMAC(OctetstringValue p_key, OctetstringValue p_message, EnumeratedValue p_crypto ) {
	    OctetstringValue ret = (OctetstringValue) RB.getTciCDRequired().getOctetstring().newInstance();
	    return ret;
	}

	/** This external function calculates a SHA1 hash
	 * @param  p_message Octetstring message
	 * @param  p_crypto Cryptographic function used to compute MAC
	 * @return Message hash
	*/
	public synchronized OctetstringValue fx_computeHash(OctetstringValue p_key, OctetstringValue p_message, EnumeratedValue p_crypto ) {
	    OctetstringValue ret = (OctetstringValue) RB.getTciCDRequired().getOctetstring().newInstance();
	    return ret;
	}

	/** This external function computes the Authenticator for 
	 *          return routability procedure
	 * @param 	p_careOfAddr Care-of address
	 * @param 	p_cnAddr Address of the correspondant
	 * @param 	p_ipv6Packet Ipv6 Packet
	 * @param 	p_kbm Binding management key
	 * @return 	Authenticator
	*/
	public synchronized OctetstringValue fx_bindingAuthenticator(OctetstringValue p_careOfAddrHae, 
			OctetstringValue p_careOfAddr, 
			OctetstringValue cnAddr,
			OctetstringValue p_ipv6Packet,
			OctetstringValue p_kbm) {
	    OctetstringValue ret = (OctetstringValue) RB.getTciCDRequired().getOctetstring().newInstance();
	    return ret;
	}
	
	/** This external function computes Integrity Check Value
	 * @param 	p_ipv6Packet Ipv6 Packet
	 * @param 	p_sa Security association
	 * @return 	Integrity Check Value
	*/
	public synchronized OctetstringValue fx_IntegrityCheckValue(UnionValue p_ipv6Packet,
			OctetstringValue p_sa) {
	    OctetstringValue ret = (OctetstringValue) RB.getTciCDRequired().getOctetstring().newInstance();
	    return ret;
	}

	/**
	 * This external function translates an IPv6 address string in its octetstring representation.
	 * @param p_address
	 * @return
	 */
	public synchronized OctetstringValue fx_translateIpv6Address(CharstringValue p_address) {
	    InetAddress addr;
	    byte bmsg[] = new byte[]{};
	    String pAddr = p_address.getString();
	    
	    if (pAddr.equals("")) {
            RB.getTciTMProvided().tciError("fx_translateIpv6Address: Cannot translate an empty string for " + p_address.getType().getName());
	    }
	    
	    try {
	    	addr = Inet6Address.getByName(pAddr);
	    	bmsg = addr.getAddress();
	    } catch (UnknownHostException e) {
	    	System.out.println("Unknown Host: " + e.getMessage());
	    }
	    
		StringBuffer sb = new StringBuffer();
		
		for (int j = 0; j < bmsg.length; j++) {
		    //we choose capital letters
		    sb.append(hex[1].charAt((bmsg[j] >> 4) & 0xf));
		    sb.append(hex[1].charAt((bmsg[j]) & 0xf));
		}		

		OctetstringValue ret = (OctetstringValue) RB.getTciCDRequired().getTypeForName("LibIpv6_MultiRfcs_TypesAndValues.Ipv6Address").newInstance();
	    if (bmsg.length != 0) {
	    	ret.setString("'" + sb + "'O");
	    }
	    else {
	    	ret.setString("'00000000000000000000000000000000'O");
	    }
	    ret.setLength(16);
	    
	    return ret;		
	}
}
